from copy import deepcopy
from datetime import datetime
import pathlib
import os
import PySimpleGUI as sg
from typing import Optional, Union, Callable
from .uploader import Uploader, utils, RetryConfig, _DEFAULT_RETRY_CONFIG
from .layout import create_layout
from .__version__ import __version__
from . import config, auth
from loguru import logger
import sys


class App:
    def __init__(
        self,
        key: Optional[Union[str, bytes]],
        title: str = f"Google Cloud Storage Uploader - {__version__}",
        with_authentication: bool = True,
        with_service_account: bool = True,
        with_bucket_id: bool = True,
        with_password: bool = False,
        with_checkboxes: bool = False,
        dry_run: bool = False,
        service_account_path: pathlib.Path = config.service_account_path,
        client_secrets_path: pathlib.Path = config.client_secrets_path,
        user_credentials_path: pathlib.Path = config.get_user_credentials_path(),
        log_file_path: pathlib.Path = config.get_log_file_path(),
        icon: Optional[pathlib.Path] = config.get_icon_path(),
        prefix: Optional[Union[str, Callable[[], str]]] = None,
        retry_config: Optional[RetryConfig] = _DEFAULT_RETRY_CONFIG,
    ) -> None:

        logger.info(f"Starting app (version {__version__})")

        self.last_updated = datetime.now()
        self.key = key
        self.dry_run = dry_run
        self.service_account_path = pathlib.Path(service_account_path)
        self.client_secrets_path = pathlib.Path(client_secrets_path)
        self.user_credentials_path = pathlib.Path(user_credentials_path)
        self.log_file_path = log_file_path
        self._prefix = prefix
        self.retry_config = retry_config

        self.default_bucket = auth.default_bucket_from_service_account_file(
            self.service_account_path
        )
        self.window = sg.Window(
            title,
            create_layout(
                with_bucket_id,
                with_authentication,
                with_service_account,
                with_checkboxes,
                with_password,
                default_bucket=self.default_bucket,
            ),
            icon=icon,  # Yes, icon is needed here!
        )
        # Hack to remove icon from title bar, but is needed above on macos!
        if sys.platform == "darwin" and icon is not None:
            self.window.set_icon(None)

    @property
    def prefix(self):
        if self._prefix is None:
            return ""
        if callable(self._prefix):
            return self._prefix()
        return self._prefix

    def callback(self, stats):
        if stats.complete:
            self.window.write_event_value("UPLOADER-DONE", deepcopy(stats))
            return

        # Update progress at a lower rate
        now = datetime.now()
        if (now - self.last_updated).seconds > 0.3:
            self.last_updated = now
            self.window.write_event_value("UPLOADER-PROGRESS", deepcopy(stats))

    def update_progress(self, stats):

        perc = (
            int(100 * (stats.bytes_so_far / stats.total_bytes))
            if stats.total_bytes
            else 0
        )
        self.window["PROGRESS"].update(stats.bytes_so_far, stats.total_bytes, True)

        self.window["PROGRESS-TEXT"].update(
            f"{utils.display_bytes(stats.bytes_so_far)} of "
            f"{utils.display_bytes(stats.total_bytes)} ({perc}%)",
        )

        msg = [
            f"Uploaded: {stats.num_uploaded} of {stats.num_files} files",
            f"Skipped: {stats.num_skipped}",
            f"Errors: {stats.num_errors}",
            f"Elapsed time: {utils.display_time(stats.elapsed_time)}",
        ]

        if stats.complete:
            msg.append("Status: Inactive")
            self.window["START"].update(disabled=False)
            self.window["STOP"].update(disabled=True)
            self.window["AUTH"].update(disabled=False)
        else:
            msg.append("Status: Active")

        self.window["STATS"].update("\n".join(msg))

    def clear_status(self):
        self.window["STATS"].update("")
        self.window["PROGRESS-TEXT"].update("")
        self.window["PROGRESS"].update(0, 100, True)

    def display_error(self, msg=None):
        msg = "View the log file for details" if msg is None else msg
        msg = f"Error occurred:\n{msg}"
        self.window["FEEDBACK"].update(msg)
        self.clear_status()

    def run(self):

        uploader = None

        while True:
            event, values = self.window.read()

            if event == sg.WIN_CLOSED:
                if uploader is not None:
                    uploader.stop()
                break

            if event == "START":
                logger.info("Action: New upload started")

                # If SA file provided, use that over user credentials
                if os.path.isfile(values.get("SA-FILE", "")):
                    user_credentials_path = None
                    sa_file = values.get("SA-FILE")
                else:
                    user_credentials_path = self.user_credentials_path
                    sa_file = self.service_account_path
                if values.get("PW"):
                    key = auth.key_from_password(values.get("PW"), self.key)
                else:
                    key = self.key

                try:
                    uploader = Uploader(
                        bucket_id=values.get("BUCKET", self.default_bucket),
                        service_account_file=sa_file,
                        user_credentials_path=user_credentials_path,
                        key=key,
                        client_secrets_path=self.client_secrets_path,
                        dry_run=self.dry_run,
                        prefix=(
                            datetime.utcnow().isoformat()
                            if values.get("DATETIME-PREFIX")
                            else self.prefix
                        ),
                        delete_files_after_upload=values.get("DELETE", False),
                        retry_confg=self.retry_config,
                    )
                    uploader.start(
                        [
                            pathlib.Path(_)
                            for _ in values["PATHS"].split(sg.BROWSE_FILES_DELIMITER)
                        ],
                        self.callback,
                    )
                    self.last_updated = datetime.now()
                    dt = self.last_updated.strftime("%d, %b at %T")
                    self.window["FEEDBACK"].update(f"Upload started on {dt}")
                    self.window["START"].update(disabled=True)
                    self.window["STOP"].update(disabled=False)
                    self.window["AUTH"].update(disabled=True)
                    self.window["PROGRESS"].update(0, 100, True)
                except ValueError as e:
                    logger.exception("Value error during startup")
                    self.display_error(msg=e)
                    continue
                except Exception:
                    logger.exception("Uploader raised exception on startup")
                    self.display_error()
                    continue

            elif event == "STOP" and uploader is not None:
                logger.info("Action: stop current upload")
                self.window["STOP"].update(disabled=True)
                uploader.stop()
                self.window["FEEDBACK"].update("Stopping transfer...")

            elif event == "AUTH":
                logger.info("Action: execute user authentication")
                Uploader.run_oauth_flow(
                    self.user_credentials_path,
                    self.client_secrets_path,
                    self.key,
                )

            elif event == "VIEW-LOG":
                logger.info("Action: View log file")
                try:
                    utils.open(self.log_file_path)
                except Exception as e:
                    logger.exception(f"Failed to open {self.log_file_path}")
                    self.window["FEEDBACK"].update(e)
                    pass
            elif event == "UPLOADER-PROGRESS":
                self.update_progress(values.get("UPLOADER-PROGRESS"))
            elif event == "UPLOADER-DONE":
                self.update_progress(values.get("UPLOADER-DONE"))
                self.window["FEEDBACK"].update("")

        self.window.close()



import math
import time
import random
import datetime
import shutil
import platform
import subprocess
import glob
import pathlib


def _find_parent(path):
    """
    This is to extract the parent directory of a wildcard without causing
    syntax issues on Windows
    """
    parent = pathlib.Path("")
    for part in path.parts:
        tmp = parent / part
        try:
            if not tmp.exists():
                break
        except Exception:
            break
        parent = tmp
        if parent == path:
            parent = path.parent
            break
    return parent


def parse_paths(paths, recursive=True):
    """
    Iterates over all file paths found in `paths`. The iterator yields a tuple
    of two elements, described as follows.

    For a given file, the first element of the tuple is the file path and the
    second is the file path relative to the parent of the specified path.

    For example,
    parse_paths('parent/dir-with-sub-dirs', True) might yield the following tuples

    (pathlib.Path('parent/dir-with-sub-dirs/f1.txt'),
     pathlib.Path('dir-with-sub-dirs/f1.txt'))

    and

    (pathlib.Path('parent/dir-with-sub-dirs/sub-dir/f2.txt'),
    pathlib.Path('dir-with-sub-dirs/sub-dir/f1.txt'))

    and `parse_paths('dir-with-sub-idirs/f1.txt')` would yield
    (pathlib.Path('dir-with-sub-dirs/f1.txt'), pathlib.Path('f1.txt'))
    """
    if isinstance(paths, str):
        paths = [paths]
    for p in paths:
        path = pathlib.Path(p).expanduser()
        path_str = str(path)
        parent = path.parent
        if path.is_dir():
            path_str = f"{path}/**"
        for _ in glob.iglob(path_str, recursive=recursive):
            path2 = pathlib.Path(_)
            if path2.is_file():
                yield (path2, path2.relative_to(parent))


def open(filepath):
    check = True
    app = None
    if platform.system() == "Windows":
        app = "explorer.exe"
        # explorer returns non-zero regardless..
        check = False
    elif platform.system() == "Darwin":
        app = "open"
    elif platform.system() == "Linux":
        app = "xdg-open"

    if app is None or not shutil.which(app):
        raise OSError(f"No known default application for opening {filepath}")
    if not filepath.is_file():
        raise ValueError(f"{filepath} does not exist")
    try:
        subprocess.run([app, str(filepath)], check=check)
    except subprocess.CalledProcessError:
        raise OSError(f"Error trying to open {filepath}")
    except OSError:
        raise OSError("No known default application")


def display_bytes(size):
    units = ["B", "KB", "MB", "GB", "TB"]
    i = 0
    while size >= 1024 and i < (len(units) - 1):
        size = size / 1024
        i += 1
    return f"{size:.2f} {units[i]}"


def display_time(seconds):
    return datetime.timedelta(seconds=int(seconds))


def path_or_partial_path_exists(path):
    path = pathlib.Path(path)
    try:
        if path.exists():
            return True
    except Exception:
        pass

    parent = pathlib.Path("")
    for part in path.parts:
        tmp = parent / part
        try:
            if tmp.exists():
                return True
        except Exception:
            pass
        parent = tmp
    return False


def backoff(n, min_backoff, max_backoff):
    min_backoff = max(min_backoff, 1)
    return min(
        (2 ** (math.log2(min_backoff) + (n - 1))) + random.uniform(0, 1),
        max_backoff,
    )


def retry_upon_error(max_retries, min_backoff, max_backoff, exceptions):
    def decorator(func):
        def inner(*args, **kwargs):
            num_retries = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    if num_retries == max_retries:
                        raise e
                    num_retries += 1
                    time.sleep(backoff(num_retries, min_backoff, max_backoff))

        return inner

    return decorator


def retry_func(func, max_retries=0, min_backoff=0, max_backoff=0, exceptions=Exception):
    num_retries = 0
    while True:
        try:
            return func()
        except exceptions as e:
            if num_retries == max_retries:
                raise e
            num_retries += 1
            time.sleep(backoff(num_retries, min_backoff, max_backoff))



import json
import click
from .uploader import Uploader
from .app import App
from . import config, auth


@click.group()
def main():
    pass


@main.command()
@click.argument("path")
@click.argument("bucket_id", type=str)
@click.option("--service-account-json", type=click.Path(exists=False))
def upload(path, bucket_id, service_account_json):
    u = Uploader(
        bucket_id=bucket_id,
        service_account_json=service_account_json,
    )
    u.start(path)


@main.command()
@click.option("--client-secrets-json", type=click.Path(exists=False), default=None)
def authenticate(client_secrets_json):
    json = client_secrets_json if client_secrets_json else config.client_secrets_path
    credentials = auth.get_credentials(json)
    auth.save_credentials(credentials, config.user_credentials_path)


@main.command()
@click.argument("path")
def gen_key(path):
    key = auth.gen_key()
    with open(path, "wb") as f:
        f.write(key)

@main.command()
@click.argument("salt-path")
@click.argument("path")
def gen_key_from_password(salt_path, path):
    pw = input("Enter password: ")
    new_key = auth.key_from_password(pw.encode('utf-8'), salt_path)
    with open(path, "wb") as f:
        f.write(new_key)

@main.command()
@click.argument("key1")
@click.argument("key2")
@click.argument("path")
def encrypt_key(key1, key2, path):
    """
    Encrypt key1 with key2
    """
    cipher = auth.get_cipher(key1)
    with open(path, "wb") as f1, open(key2, "rb") as f2:
        f1.write(cipher.encrypt(f2.read()))

@main.command("pickle-service-account-file")
@click.argument("path-in")
@click.argument("key-file")
@click.option("--output-file", default=config.service_account_path)
@click.option("--bucket", default=None)
def pickle_service_account_file(path_in, key_file, output_file, bucket):
    with open(path_in, "r") as f:
        data = json.load(f)
    auth.pickle_service_account_credentials(data, output_file, key_file, bucket)


@main.command("pickle-client-secrets-file")
@click.argument("path-in")
@click.argument("key-file")
def pickle_client_secrets_file(path_in, key_file):
    with open(path_in, "r") as f:
        data = json.load(f)
    auth.pickle_client_secret_file(data, config.client_secrets_path, key_file)


@main.command()
@click.option("--key-file", type=click.Path(exists=False), default=None)
@click.option(
    "--service-account-file",
    type=click.Path(exists=False),
    default=config.service_account_path,
)
@click.option(
    "--user-credentials-file",
    type=click.Path(exists=False),
    default=config.get_user_credentials_path(),
)
@click.option(
    "--client-secret-file",
    type=click.Path(exists=False),
    default=config.client_secrets_path,
)
@click.option("--title", default="GCS Uploader")
@click.option("--with-authentication", is_flag=True)
@click.option("--with-service-account", is_flag=True)
@click.option("--with-bucket-id", is_flag=True)
@click.option("--with-password", is_flag=True)
@click.option("--with-checkboxes", is_flag=True)
@click.option("--dry-run", is_flag=True)
def gui(
    key_file,
    service_account_file,
    user_credentials_file,
    client_secret_file,
    title,
    with_authentication,
    with_service_account,
    with_bucket_id,
    with_checkboxes,
    with_password,
    dry_run,
):
    app = App(
        key=key_file,
        title=title,
        with_authentication=with_authentication,
        with_service_account=with_service_account,
        with_bucket_id=with_bucket_id,
        with_checkboxes=with_checkboxes,
        with_password=with_password,
        dry_run=dry_run,
        service_account_path=service_account_file,
        user_credentials_path=user_credentials_file,
        client_secrets_path=client_secret_file,
    )
    app.run()



from dataclasses import dataclass, field
import os
import threading
import pathlib
import time
from google.cloud import storage
import google.auth.exceptions
import requests.exceptions
import concurrent.futures as cf
from loguru import logger
from typing import Optional, Union, Tuple, Type
from . import auth
from . import exceptions
from . import utils


@dataclass
class RetryConfig:
    num_retries: int
    max_backoff: int
    min_backoff: int
    exceptions: Optional[Tuple[Type[Exception]]] = None


_DEFAULT_RETRY_CONFIG = RetryConfig(
    num_retries=3,
    max_backoff=120,
    min_backoff=20,
    exceptions=(
        google.auth.exceptions.TransportError,
        requests.exceptions.ConnectionError,
        requests.exceptions.ChunkedEncodingError,
    ),
)


@dataclass
class UploadStatus:
    filepath: pathlib.Path
    time_started: float = field(default_factory=time.time)
    time_complete: float = 0
    size: int = 0
    bytes_so_far: int = 0
    total_bytes: int = 0
    num_uploaded: int = 0
    num_files: int = 0
    num_errors: int = 0
    num_skipped: int = 0
    complete: bool = False

    def done(self) -> None:
        self.time_complete = time.time()
        self.complete = True

    @property
    def elapsed_time(self) -> float:
        return time.time() - self.time_started


class Uploader:
    def __init__(
        self,
        bucket_id: Optional[str],
        service_account_file: Optional[pathlib.Path],
        user_credentials_path: Optional[pathlib.Path],
        key: Optional[Union[str, bytes]],
        client_secrets_path: Optional[pathlib.Path],
        dry_run: bool,
        prefix: Optional[str],
        delete_files_after_upload: bool,
        retry_confg: Optional[RetryConfig] = _DEFAULT_RETRY_CONFIG,
    ) -> None:

        self._stop = False
        self._thread = None
        self.dry_run = dry_run
        self.futures = None
        self.delete_files_after_upload = delete_files_after_upload
        self.lock = threading.Lock()
        if retry_confg is None:
            self._retry_config = RetryConfig(
                num_retries=0, max_backoff=0, min_backoff=0, exceptions=(Exception,)
            )
        else:
            self._retry_config = retry_confg
        if prefix:
            self.prefix = prefix.rstrip("/") + "/"
        else:
            self.prefix = ""

        if bucket_id is not None and len(bucket_id) < 3:
            raise ValueError("Must provide a valid bucket ID")
        self.bucket_id = bucket_id

        self.client = None

        service_account_file = (
            pathlib.Path(service_account_file)
            if service_account_file
            else pathlib.Path("")
        )

        credentials = None

        service_account_file = (
            pathlib.Path(service_account_file)
            if service_account_file is not None
            else None
        )
        user_credentials_path = (
            pathlib.Path(user_credentials_path)
            if user_credentials_path is not None
            else None
        )
        client_secrets_path = (
            pathlib.Path(client_secrets_path)
            if client_secrets_path is not None
            else None
        )
        try:
            if (
                isinstance(user_credentials_path, pathlib.Path)
                and user_credentials_path.is_file()
            ):
                logger.debug(f"Using {user_credentials_path}")
                credentials = auth.unpickle_user_credentials(user_credentials_path, key)
            elif (
                isinstance(service_account_file, pathlib.Path)
                and service_account_file.is_file()
            ):
                logger.debug(f"Using {service_account_file}")
                credentials, bucket = auth.unpickle_service_account_credentials(
                    service_account_file, key, True
                )
                self.bucket_id = bucket if self.bucket_id is None else self.bucket_id
            elif (
                isinstance(client_secrets_path, pathlib.Path)
                and client_secrets_path.is_file()
            ):
                logger.debug(f"Running auth flow with {client_secrets_path}")
                credentials = self.run_oauth_flow(
                    user_credentials_path, client_secrets_path, key
                )
            self.client = storage.Client(project=None, credentials=credentials)
        except Exception:
            logger.exception("Authentication error")
            raise ValueError("Authentication failed")

    @staticmethod
    def run_oauth_flow(user_credentials_path, client_secrets_path, key):
        credentials = auth.get_credentials_via_client_secrets_file(
            client_secrets_path, key
        )
        auth.pickle_user_credentials(credentials, user_credentials_path, key)
        return credentials

    def start(
        self,
        paths,
        callback=lambda x: None,
    ):

        if not (paths and all([pathlib.Path(_).exists() for _ in paths])):
            raise ValueError("Must provide valid paths to files or folders")

        self._stop = False

        try:
            bucket = self.client.bucket(self.bucket_id)
        except Exception:
            logger.exception("Failed to instantiate bucket")
            raise Exception("Failed to connect as invalid destination URI")

        self._thread = threading.Thread(
            target=self._future_handler,
            args=(bucket, paths, callback),
            daemon=True,
        )
        self._thread.start()

    def _future_handler(self, bucket, paths, callback):

        self.futures = {}

        job = self._mock_upload if self.dry_run else self._upload

        self._upload_status = UploadStatus(None)

        max_workers = max(os.cpu_count() - 1, 2)
        with cf.ThreadPoolExecutor(max_workers=max_workers) as executor:

            for path, dest in utils.parse_paths(paths):
                if self._stop:
                    break

                size = path.stat().st_size

                # Start accumulating sizes for running stats
                with self.lock:
                    self._upload_status.total_bytes += size
                    self._upload_status.num_files += 1

                self.futures[
                    executor.submit(
                        job,
                        path,
                        (pathlib.Path(self.prefix) / dest).as_posix(),
                        bucket,
                        callback,
                    )
                ] = path

        self._upload_status.done()
        callback(self._upload_status)

    @property
    def active(self):
        if self._thread is not None and self._thread.is_alive():
            return True
        return False

    def stop(self):
        self._stop = True
        if self.futures:
            [_.cancel() for _ in self.futures.keys() if not _.done()]

    def _upload_with_retry(self, path, dest, bucket):
        blob = bucket.blob(dest)
        input("Waiting for input")
        logger.info(f"Attempting to upload {path} to {dest}")
        return utils.retry_func(
            lambda: blob.upload_from_filename(path, checksum="crc32c"),
            max_retries=self._retry_config.num_retries,
            min_backoff=self._retry_config.min_backoff,
            max_backoff=self._retry_config.max_backoff,
            exceptions=self._retry_config.exceptions,
        )

    def _upload(self, path, dest, bucket, cb):
        if self._stop:
            return
        skipped = num_errors = False
        try:
            self._upload_with_retry(str(path), str(dest), bucket)
        except Exception as e:
            if exceptions.is_delete_permission_error(e):
                logger.warning(f"{path} already exists on the remote")
                skipped = True
            else:
                logger.exception(f"Error attempting uploading {path}")
                num_errors = True

        with self.lock:
            if skipped or num_errors:
                self._upload_status.num_skipped += 1
                self._upload_status.num_errors += 1
            else:
                self._upload_status.num_uploaded += 1
                self._upload_status.bytes_so_far += path.stat().st_size
                if self.delete_files_after_upload:
                    logger.info(f"Deleting {path}")
                    path.unlink()
            self._upload_status.filepath = path
            cb(self._upload_status)

    def _mock_upload(self, path, dest, bucket, cb):
        if self._stop:
            return
        time.sleep(1)
        with self.lock:
            self._upload_status.num_uploaded += 1
            self._upload_status.bytes_so_far += path.stat().st_size
            cb(self._upload_status)



import re
import google.api_core.exceptions as gae

e_msg_regex = re.compile(r'message":\s?"(.*?)"', re.IGNORECASE)


def is_delete_permission_error(e, filename=None):
    if not isinstance(e, gae.Forbidden):
        return False
    try:
        msg = e_msg_regex.search(str(e)).group(1)
        if "storage.objects.delete" in msg:
            return True
    except Exception:
        return False


class UploadError(Exception):
    def __init__(self, filename, message=""):
        self.err = f"Failed to upload {filename}: {message}"
        super().__init__(self.err)


class UploadOverwriteError(UploadError):
    def __init__(self, filename, message="File already exists on remote"):
        super().__init__(filename, message=message)


class UploadCancelled(UploadError):
    def __init__(self, filename, message="Upload cancelled"):
        super().__init__(filename, message=message)



import os
import sys
import pathlib
from appdirs import AppDirs
from typing import Optional

# Path stuff
bundle_dir = pathlib.Path(getattr(sys, "_MEIPASS", "."))
client_secrets_path = bundle_dir / "data/client_secrets.pickle"
service_account_path = bundle_dir / "data/service_account.pickle"


def get_icon_path() -> Optional[pathlib.Path]:
    if os.name == "nt":
        icon_path = bundle_dir / "data/icon.ico"
    else:
        icon_path = bundle_dir / "data/icon.icns"
    if icon_path.is_file():
        return icon_path.absolute()
    return None


dirs = AppDirs("bucketpusher")
user_data_dir = pathlib.Path(dirs.user_data_dir)


def get_user_credentials_path() -> pathlib.Path:
    user_credentials_path = user_data_dir / "user_credentials.pickle"
    if not user_credentials_path.parent.is_dir():
        user_data_dir.parent.mkdir(parents=True, exist_ok=True)
    return user_credentials_path


def get_log_file_path() -> pathlib.Path:
    log_file = pathlib.Path(dirs.user_log_dir) / "bucketpusher.log"

    if not log_file.parent.is_dir():
        log_file.parent.mkdir(parents=True, exist_ok=True)
    return log_file



from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64
import pickle
import json
import os
import pathlib
from typing import Dict, Any, Union, Iterable, Optional, Tuple
from google_auth_oauthlib import flow
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials

KeyT = Union[str, bytes, pathlib.Path]


def _load_file(path: pathlib.Path) -> Dict[str, Any]:
    if path.suffix == ".json":
        with open(path, "rb") as f:
            data = json.load(f)
    elif path.suffix == ".pickle":
        with open(path, "rb") as f:
            data = pickle.load(f)
    else:
        raise ValueError(f"Unknown file type: {path}")
    return data


def get_credentials(client_id: str, client_secret: str) -> Credentials:
    client_config = {
        "installed": {
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob", "http://localhost"],
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
        }
    }
    scopes = ("https://www.googleapis.com/auth/devstorage.read_write",)

    app_flow = flow.InstalledAppFlow.from_client_config(client_config, scopes=scopes)
    app_flow.run_local_server()  # Should catch exception here?
    return app_flow.credentials


def credentials_to_dict(credentials: Credentials) -> Dict[str, Any]:
    return {
        "token": credentials.token,
        "refresh_token": credentials.refresh_token,
        "token_uri": credentials.token_uri,
        "client_id": credentials.client_id,
        "client_secret": credentials.client_secret,
        "scopes": credentials.scopes,
    }


def pickle_user_credentials(
    credentials: Credentials,
    path: pathlib.Path,
    key: KeyT,
) -> None:
    path = pathlib.Path(path)
    if not path.parent.is_dir():
        path.parent.mkdir(parents=True, exist_ok=True)
    credentials.token = None
    data = credentials_to_dict(credentials)
    encrypt_fields(data, ("refresh_token", "client_id", "client_secret"), key)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def unpickle_user_credentials(path: pathlib.Path, key: KeyT) -> Credentials:
    data = _load_file(path)
    decrypt_fields(data, ("refresh_token", "client_id", "client_secret"), key)
    return Credentials(**data)


def key_from_password(pw, salt):
    if os.path.isfile(salt):
        with open(salt, "rb") as f:
            salt = f.read()
    elif isinstance(salt, str):
        salt = salt.encode('utf-8')
    if isinstance(pw, str):
        pw = pw.encode('utf-8')
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
    )
    key = base64.urlsafe_b64encode(kdf.derive(pw))
    return key

def encrypt_fields(data: Dict[str, Any], keys: Iterable[str], key: KeyT) -> None:
    cipher = get_cipher(key)
    for k in keys:
        data[k] = cipher.encrypt(data[k].encode("utf-8"))


def decrypt_fields(data: Dict[str, Any], keys: Iterable[str], key: KeyT) -> None:
    if key is None:
        return
    cipher = get_cipher(key)
    for k in keys:
        data[k] = cipher.decrypt(data[k]).decode("utf-8")


def unpickle_service_account_credentials(
    path: pathlib.Path, key: KeyT, as_credentials: bool = True
) -> Tuple[Union[Credentials, Dict[str, Any]], Optional[str]]:
    data = _load_file(path)
    decrypt_fields(data, ("private_key_id", "private_key"), key)
    bucket = data.pop("bucket") if "bucket" in data else None
    if as_credentials:
        creds: Credentials = service_account.Credentials.from_service_account_info(data)
        return creds, bucket
    return data, bucket


def pickle_service_account_credentials(
    data: Dict[str, Any], path: pathlib.Path, key: KeyT, bucket: Optional[str] = None
) -> None:
    data = {
        k: data[k]
        for k in [
            "private_key",
            "private_key_id",
            "token_uri",
            "client_email",
        ]
    }
    encrypt_fields(data, ("private_key_id", "private_key"), key)
    if bucket is not None:
        data["bucket"] = bucket
    with open(path, "wb") as f:
        pickle.dump(data, f)


def unpickle_client_secret_file(path: pathlib.Path, key: KeyT) -> Dict[str, Any]:
    with open(path, "rb") as f:
        data: Dict[str, Any] = pickle.load(f)
    decrypt_fields(
        data["installed"],
        ("client_secret",),
        key,
    )
    return data


def pickle_client_secret_file(
    data: Dict[str, Any], path: pathlib.Path, key: KeyT
) -> None:
    encrypt_fields(data["installed"], ("client_secret",), key)
    with open(path, "wb") as f:
        pickle.dump(data, f)


def get_credentials_via_client_secrets_file(
    pickled_client_secret_file: pathlib.Path, key: KeyT
) -> Credentials:
    data = unpickle_client_secret_file(pickled_client_secret_file, key)
    appflow = flow.InstalledAppFlow.from_client_config(
        data,
        scopes=["https://www.googleapis.com/auth/devstorage.read_write"],
    )
    appflow.run_local_server()
    return appflow.credentials


def gen_key() -> bytes:
    return Fernet.generate_key()


def get_cipher(key: KeyT) -> Fernet:
    if os.path.isfile(key):
        with open(key, "rb") as f:
            return Fernet(f.read())
    elif isinstance(key, (str, bytes)):
        return Fernet(key)
    else:
        raise ValueError(f"Failed to instantiate new `Fernet` from {key}")


def default_bucket_from_service_account_file(
    service_account_path: pathlib.Path,
) -> Optional[str]:
    if not os.path.isfile(service_account_path):
        return None
    data = _load_file(service_account_path)
    return data.pop("bucket") if "bucket" in data else None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bucketpusher']

package_data = \
{'': ['*']}

install_requires = \
['PySimpleGUI>=4.30.0,<5.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'cryptography>=3.2.1,<4.0.0',
 'flake8>=3.8.4,<4.0.0',
 'google-auth-oauthlib>=0.4.2,<0.5.0',
 'google-cloud-storage>=1.32.0,<2.0.0',
 'google-crc32c>=1.0.0,<2.0.0',
 'loguru>=0.5.3,<0.6.0']

entry_points = \
{'console_scripts': ['bucketpusher = bucketpusher:cli.main']}

setup_kwargs = {
    'name': 'bucketpusher',
    'version': '0.4.0',
    'description': 'Push local files to cloud storage bucket',
    'long_description': None,
    'author': 'Dominic Ward',
    'author_email': 'dom@deeuu.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pathlib
import PySimpleGUI as sg

sg.theme("Reddit")


def create_progress_status():
    return sg.Column(
        [
            [
                sg.Frame(
                    layout=[
                        [
                            sg.ProgressBar(
                                100, orientation="h", key="PROGRESS", size=(50, 2)
                            )
                        ],
                        [
                            sg.Text(
                                key="PROGRESS-TEXT",
                                justification="center",
                                size=(60, 1),
                                text_color="light blue",
                            )
                        ],
                        [
                            sg.Text(
                                key="STATS",
                                justification="left",
                                size=(30, 5),
                                text_color="#b4c1c5",
                            )
                        ],
                        [
                            sg.Text(
                                key="FEEDBACK",
                                justification="center",
                                size=(60, 2),
                                text_color="#f39a87",
                            )
                        ],
                    ],
                    title="Status",
                    element_justification="center",
                )
            ]
        ],
        justification="center",
    )


def create_bottom_buttons(with_authentication=True):

    buttons = [
        sg.Button(
            "Start",
            key="START",
            size=(10, 1),
        ),
        sg.Button("Stop", key="STOP", size=(10, 1), disabled=True),
        sg.Button("View log", key="VIEW-LOG", size=(10, 1), disabled=False),
        sg.Button(
            "Authenticate", key="AUTH", size=(10, 1), visible=with_authentication
        ),
    ]

    bottom_buttons = sg.Column([buttons], justification="center")
    return bottom_buttons


def create_checkboxes():
    result = sg.Column(
        [
            [
                sg.Checkbox("Delete after upload", default=False, key="DELETE"),
                sg.Checkbox(
                    "Date & time as prefix", default=False, key="DATETIME-PREFIX"
                ),
            ]
        ],
        justification="center",
    )
    return result


def create_layout(
    with_bucket_id=True,
    with_authentication=True,
    with_service_account=True,
    with_checkboxes=True,
    with_key=True,
    default_file_browser_text=(
        f"e.g. '{pathlib.Path.home() / 'directory-to-upload'}'   or browse →"
    ),
    default_bucket="",
):

    layout = [
        [
            sg.Text("File(s) or Folder", size=(20, 1)),
            sg.Input(
                default_file_browser_text,
                key="PATHS",
            ),
            sg.FilesBrowse("Files", target="PATHS"),
            sg.FolderBrowse("Folder", target="PATHS"),
        ],
    ]
    if with_bucket_id:
        layout.append(
            [
                sg.Text("Bucket ID", size=(20, 1)),
                sg.Input(default_bucket, key="BUCKET"),
            ],
        )
    if with_service_account:
        layout.append(
            [
                sg.Text("Service Account File", size=(20, 1)),
                sg.Input(
                    "",
                    key="SA-FILE",
                ),
                sg.FileBrowse(),
            ]
        )
    if with_key:
        layout.append(
            [
                sg.Text("Password", size=(20, 1)),
                sg.Input(
                    "",
                    key="PW",
                    password_char='*',
                ),
            ]
        )

    if with_checkboxes:
        layout += [[create_checkboxes()]]

    layout += [
        [create_progress_status()],
        [create_bottom_buttons(with_authentication)],
    ]

    return layout
